from __future__ import absolute_import

import numpy as np
from obspy.core.utcdatetime import UTCDateTime
from obspy import Trace
from typing import List, Optional

from .MetadataAlgorithm import MetadataAlgorithm
from ..metadata import Metadata
from ..metadata.flag.Flag import ArtificialDisturbance, ArtificialDisturbanceType
from ..metadata import MetadataCategory


class SpikesAlgorithm(MetadataAlgorithm):
    """
    Algorithm that identifies spikes in data and generates metadata flags.

    Attributes
    ----------
    window_size: int
        Size of the rolling window for dynamic alignment (no shrinking or padding). Defaults to 1000.
    threshold: int
        Threshold for spike detection (threshold * sigma). Defaults to 6.
    """

    window_size: Optional[int] = 1000
    threshold: Optional[int] = 6

    def _despike(
        self, signal: Trace, timestamps: List[UTCDateTime]
    ) -> List[UTCDateTime]:
        """
        Internal method to compute spike timestamps using a rolling window with dynamic alignment.

        Parameters
        ----------
        signal: Trace
            Signal trace data to analyze.
        timestamps: List[UTCDateTime]
            Corresponding timestamps for the signal data.

        Returns
        -------
        List[UTCDateTime]
            Timestamps where spikes were detected.
        """
        n = len(signal)
        half_window = self.window_size // 2

        rolling_mean = np.zeros(n)
        rolling_std = np.zeros(n)

        for i in range(n):
            start = max(0, i - half_window)
            end = min(n, i + half_window + 1)

            window = signal[start:end]
            rolling_mean[i] = np.mean(window)
            rolling_std[i] = np.std(window)

        spikes = np.abs(signal - rolling_mean) > self.threshold * rolling_std
        return timestamps[spikes]

    def run(self) -> List[ArtificialDisturbance]:
        """
        Run the despiking algorithm and generate spike metadata.

        Returns
        -------
        List[ArtificialDisturbance]
            List of metadata objects for detected spikes.
        """
        stream = self.get_stream()
        daily_streams = self.split_stream_by_day(stream)
        spike_metadata_list = []

        for day_stream in daily_streams:
            for trace in day_stream:
                spike_metadata = self.process_trace(trace)
                if spike_metadata:
                    spike_metadata_list.append(spike_metadata)

        return spike_metadata_list

    def process_trace(self, trace: Trace) -> Optional[ArtificialDisturbance]:
        """
        Process a single trace, detecting spikes and creating or updating metadata as needed.

        Parameters
        ----------
        trace : Trace
            The trace to process for spikes.

        Returns
        -------
        Optional[ArtificialDisturbance]
            The metadata for the trace if spikes are detected, None otherwise.
        """

        timestamps = trace.times("UTCDateTime")
        signal = trace.data
        # TODO: throw error if singal len < window!
        spike_timestamps = self._despike(signal, timestamps)

        # check existing metadata
        existing_metadata = self.check_existing_spike_metadata(trace)

        if len(spike_timestamps) > 0:
            spike_metadata = self.create_spike_metadata(spike_timestamps, trace)
            if existing_metadata:
                # update ID to preexisting one
                spike_metadata.id = existing_metadata[0].id
            return spike_metadata

        if len(spike_timestamps) == 0 and existing_metadata is not None:
            # TODO:change existing spike metadata to rejected or data_valid to false? Update array?
            # convert existing_metadata back into Metadata
            pass

        return None

    def check_existing_spike_metadata(self, trace: Trace) -> str | None:
        """
        Find existing spike metadata for a specified period.

        Parameters
        ----------
        trace : Trace
            The trace from which the metadata is searched.
        Returns
        -------
        The existing metadata if found, None otherwise.
        """
        spike_metadata = Metadata(
            category=MetadataCategory.FLAG,
            station=trace.stats.station,
            starttime=trace.stats.starttime,
            endtime=trace.stats.endtime,
            channel=trace.stats.channel,
        )
        return self.check_existing_metadata(spike_metadata)

    def create_spike_metadata(
        self, spike_timestamps: List[UTCDateTime], trace: Trace
    ) -> ArtificialDisturbance:
        """
        Create new metadata for detected spikes in a trace.

        Parameters
        ----------
        spike_timestamps : List[UTCDateTime]
            List of timestamps where spikes were detected.
        trace : Trace
            The trace from which the metadata is created.

        Returns
        -------
        ArtificialDisturbance
            Metadata object for the detected spikes.
        """
        metadata_dict = {
            "description": f"Spikes detected using a window of {self.window_size} and a threshold of {self.threshold}*sigma",
            "artificial_disturbance_type": ArtificialDisturbanceType.SPIKES,
            "spikes": list(spike_timestamps),
        }

        return self.create_metadata(
            starttime=spike_timestamps[0],
            endtime=spike_timestamps[-1],
            metadata_class=ArtificialDisturbance,
            metadata_dict=metadata_dict,
            category=MetadataCategory.FLAG,
            network=trace.stats.network,
            channel=trace.stats.channel,
            location=trace.stats.location,
            created_by="SpikesAlgorithm",
            status="New",
        )
