import datetime
import enum
import os
from typing import List, Optional, Union

from obspy import UTCDateTime
from pydantic import ConfigDict, field_validator, model_validator, Field, BaseModel

from .Element import ELEMENTS
from .Observatory import OBSERVATORY_INDEX, ASL_OBSERVATORY_INDEX
from ...pydantic_utcdatetime import CustomUTCDateTimeType


DEFAULT_ELEMENTS = ["X", "Y", "Z", "F"]
REQUEST_LIMIT = 3456000  # Increased the request limit by 10x what was decided by Jeremy
VALID_ELEMENTS = [e.id for e in ELEMENTS]


class DataType(str, enum.Enum):
    VARIATION = "variation"
    ADJUSTED = "adjusted"
    QUASI_DEFINITIVE = "quasi-definitive"
    DEFINITIVE = "definitive"

    @classmethod
    def values(cls) -> List[str]:
        return [t.value for t in cls]


class OutputFormat(str, enum.Enum):
    IAGA2002 = "iaga2002"
    JSON = "json"


class SamplingPeriod(float, enum.Enum):
    TEN_HERTZ = 0.1
    SECOND = 1.0
    MINUTE = 60.0
    HOUR = 3600.0
    DAY = 86400.0


class DataHost(str, enum.Enum):
    # recognized public Edge data hosts, plus one user-specified
    DEFAULT = os.getenv("DATA_HOST", "edgecwb.usgs.gov")
    EDGECWB = "edgecwb.usgs.gov"
    CWBPUB = "cwbpub.cr.usgs.gov"
    CWBPUB2 = "cwbp2.cr.usgs.gov"

    @classmethod
    def values(cls) -> List[str]:
        return [t.value for t in cls]


def default_starttime() -> UTCDateTime:
    # default to start of current day
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    return UTCDateTime(year=now.year, month=now.month, day=now.day)


class DataApiQuery(BaseModel):
    model_config = ConfigDict(extra="forbid")

    id: str
    starttime: CustomUTCDateTimeType = Field(
        default_factory=lambda: default_starttime()
    )
    # endtime default is dependent on start time, so it's handled after validation in the model_validator
    endtime: Optional[CustomUTCDateTimeType] = None
    elements: List[str] = DEFAULT_ELEMENTS
    sampling_period: SamplingPeriod = SamplingPeriod.MINUTE
    data_type: Union[DataType, str] = DataType.VARIATION
    format: Union[OutputFormat, str] = OutputFormat.IAGA2002
    data_host: Union[DataHost, str] = DataHost.DEFAULT

    @field_validator("starttime", mode="before")
    def validate_starttime(
        cls, starttime: CustomUTCDateTimeType
    ) -> CustomUTCDateTimeType:
        if not starttime:
            return default_starttime()
        else:
            return starttime

    @field_validator("data_host", mode="before")
    def validate_data_host(cls, data_host: DataHost) -> DataHost:
        if data_host not in DataHost.values():
            raise ValueError(
                # don't advertise acceptable hosts
                f"Bad data_host value '{data_host}'."
            )
        return data_host

    @field_validator("elements", mode="before")
    @classmethod
    def validate_elements(cls, elements: List[str]) -> List[str]:
        if len(elements) == 1 and "," in elements[0]:
            elements = [e.strip() for e in elements[0].split(",")]
        for element in elements:
            if element not in VALID_ELEMENTS and len(element) != 3:
                raise ValueError(
                    f"Bad element '{element}'."
                    f" Valid values are: {', '.join(VALID_ELEMENTS)}."
                )
        return elements

    @field_validator("id", mode="before")
    def validate_id(cls, id: str) -> str:
        complete_observatory_index = {**OBSERVATORY_INDEX, **ASL_OBSERVATORY_INDEX}
        if id not in complete_observatory_index:
            raise ValueError(
                f"Bad observatory id '{id}'."
                f" Valid values are: {', '.join(sorted(complete_observatory_index.keys()))}."
            )
        return id

    @model_validator(mode="after")
    def validate_combinations(self):
        if len(self.elements) > 4 and self.format == "iaga2002":
            raise ValueError("No more than four elements allowed for iaga2002 format.")
        if self.endtime is None:
            # end time defaults to 1 day after start time
            self.endtime = self.starttime + (86400 - 0.001)
        if self.starttime > self.endtime:
            raise ValueError("Starttime must be before endtime.")
        # check data volume
        samples = int(
            len(self.elements) * (self.endtime - self.starttime) / self.sampling_period
        )
        if samples > REQUEST_LIMIT:
            raise ValueError(f"Request exceeds limit ({samples} > {REQUEST_LIMIT})")
        # otherwise okay
        return self
